//-----------------------------------------------------------------------------
// File: DDEx3.CPP
//
// Desc: Direct Draw example program 3.  Adds functionality to
//       example program 2.  Creates two offscreen surfaces in
//       addition to the primary surface and back buffer.  Loads
//       a bitmap file into each offscreen surface.  Uses BltFast
//       to copy the contents of an offscreen surface to the back
//       buffer and then flips the buffers and copies the next
//       offscreen surface to the back buffer.  Press F12 to exit
//       the program.  This program requires at least 1.2 Megs of
//       video ram.
//
// Copyright (c) 1995-1999 Microsoft Corporation. All rights reserved.
// Ported by Vadim Bodrov, TMT Development Corp.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------
uses Windows, Messages, DXLite;

{$r ddex3.res}

const
    IDI_MAIN_ICON  = 101;
    IDR_MENU       = 102;
    IDM_EXIT       = 40001;


//-----------------------------------------------------------------------------
// Local definitions
//-----------------------------------------------------------------------------
const
    NAME            = 'DDExample3';
    TITLE           = 'Direct Draw Example 3';

//-----------------------------------------------------------------------------
// Default settings
//-----------------------------------------------------------------------------
const
     TIMER_ID       = 1;
     TIMER_RATE     = 500;

//-----------------------------------------------------------------------------
// Global data
//-----------------------------------------------------------------------------
var
    g_pDD:         PPDirectDraw7 := nil;          // DirectDraw object
    g_pDDSPrimary: PPDirectDrawSurface7 := nil;   // DirectDraw primary surface
    g_pDDSBack:    PPDirectDrawSurface7 := nil;   // DirectDraw back surface
    g_pDDSOne:     PPDirectDrawSurface7 := nil;   // Offscreen surface 1
    g_pDDSTwo:     PPDirectDrawSurface7 := nil;   // Offscreen surface 2
    g_pDDPal:      PPDirectDrawPalette := nil;    // The primary surface palette
    g_bActive:     Boolean := FALSE;              // Is application active?

//-----------------------------------------------------------------------------
// Local data
//-----------------------------------------------------------------------------
var
    szBitmap: PChar     := 'DDEX3';
    szMsg: PChar        := 'Page Flipping Test: Press F12 to exit';
    szFrontMsg: PChar   := 'Front buffer (F12 to quit)';
    szBackMsg: PChar    := 'Back buffer (F12 to quit)';



//-----------------------------------------------------------------------------
// Name: ReleaseAllObjects()
// Desc: Finished with all objects we use; release them
//-----------------------------------------------------------------------------
procedure ReleaseAllObjects;
begin
    if g_pDD  <> nil then
    begin
        if g_pDDSPrimary <> nil then
        begin
            g_pDDSPrimary^^.Release^(g_pDDSPrimary);
            g_pDDSPrimary := nil;
        end;
        g_pDD^^.Release^(g_pDD);
        g_pDD := nil;
    end;
end;




//-----------------------------------------------------------------------------
// Name: InitFail()
// Desc: This function is called if an initialization function fails
//-----------------------------------------------------------------------------
function InitFail(Wnd: HWND; hRet: HREsult; szError: LPCTSTR): HResult;
begin
    ReleaseAllObjects;
    MessageBox(Wnd, szError, TITLE, MB_OK);
    DestroyWindow(Wnd);
    Result := hRet;
end;


//-----------------------------------------------------------------------------
// Name: InitSurfaces()
// Desc: This function reads the bitmap file FRNTBACK.BMP and stores half of it
//       in offscreen surface 1 and the other half in offscreen surface 2.
//-----------------------------------------------------------------------------
function InitSurfaces: Boolean;
var
    hbm: HBITMAP;
begin
    // Load our bitmap resource.
    hbm := LoadImage(GetModuleHandle(nil), szBitmap, IMAGE_BITMAP, 0,
                              0, LR_CREATEDIBSECTION);
    if (hbm = 0) then
    begin
        Result := FALSE;
        exit;
    end;

    DDCopyBitmap(g_pDDSOne, hbm, 0, 0, 640, 480);
    DDCopyBitmap(g_pDDSTwo, hbm, 0, 480, 640, 480);
    DeleteObject(hbm);
    Result := TRUE;
end;



//-----------------------------------------------------------------------------
// Name: RestoreAll()
// Desc: Restore all lost objects
//-----------------------------------------------------------------------------
function RestoreAll: HResult;
var
    hRet: HResult;
begin
    hRet := g_pDDSPrimary^^.Restore^(g_pDDSPrimary);
    if (hRet = DD_OK) then
    begin
        hRet := g_pDDSOne^^.Restore^(g_pDDSOne);
        if (hRet = DD_OK) then
        begin
            hRet := g_pDDSTwo^^.Restore^(g_pDDSTwo);
            if (hRet = DD_OK) then
            begin
                InitSurfaces;
            end;
        end;
    end;
    Result := hRet;
end;




//-----------------------------------------------------------------------------
// Name: UpdateFrame()
// Desc: Displays the proper image for the page
//-----------------------------------------------------------------------------
var
    phase: Byte := 0;

procedure UpdateFrame(Wnd: HWND);
var
    hRet:   HResult;
    pdds:   PPDirectDrawSurface7;
    rcRect: TRect;
begin
    rcRect.left := 0;
    rcRect.top := 0;
    rcRect.right := 640;
    rcRect.bottom := 480;
    if (phase <> 0) then
    begin
        pdds := g_pDDSTwo;
        phase := 0;
    end else
    begin
        pdds := g_pDDSOne;
        phase := 1;
    end;
    while (TRUE) do
    begin
        hRet := g_pDDSBack^^.BltFast^(g_pDDSBack, 0, 0, pdds, rcRect, LONGFALSE);
        if (hRet = DD_OK) then
            break;
        if (hRet = DDERR_SURFACELOST) then
        begin
            hRet := RestoreAll;
            if (hRet <> DD_OK) then
                break;
        end;
        if (hRet <> DDERR_WASSTILLDRAWING) then
            break;
    end;
end;




//-----------------------------------------------------------------------------
// Name: WindowProc()
// Desc: The Main Window Procedure
//-----------------------------------------------------------------------------
function WindowProc conv arg_stdcall (Wnd: HWND; message: UINT; wParam: WPARAM; lParam: LPARAM): LResult;
var
    hRet: HResult;
begin
    case message of
        WM_ACTIVATE:
        begin
            // Pause if minimized
            g_bActive := not Boolean(HIWORD(wParam));
            Result := 0;
            exit;
        end;

        WM_DESTROY:
        begin
            // Clean up and close the app
            ReleaseAllObjects;
            PostQuitMessage(0);
            Result := 0;
            exit;
        end;

        WM_KEYDOWN:
        begin
            // Handle any non-accelerated key commands
            if wParam in [VK_ESCAPE, VK_F12] then
            begin
                    PostMessage(Wnd, WM_CLOSE, 0, 0);
                    Result := 0;
                    exit;
            end;
        end;

        WM_SETCURSOR:
        begin
            // Turn off the cursor since this is a full-screen app
            SetCursor(0);
            Result := LONGTRUE;
            exit;
        end;

        WM_TIMER:
        begin
            // Update and flip surfaces
            if (g_bActive) and (TIMER_ID = wParam) then
            begin
                UpdateFrame(Wnd);
                while (TRUE) do
                begin
                    hRet := g_pDDSPrimary^^.Flip^(g_pDDSPrimary, nil, 0);
                    if (hRet = DD_OK) then
                        break;
                    if (hRet = DDERR_SURFACELOST) then
                    begin
                        hRet := RestoreAll;
                        if (hRet <> DD_OK) then
                            break;
                    end;
                    if (hRet <> DDERR_WASSTILLDRAWING) then
                        break;
                end;
            end;
        end;
    end;
    Result := DefWindowProc(Wnd, message, wParam, lParam);
end;




//-----------------------------------------------------------------------------
// Name: InitApp()
// Desc: Do work required for every instance of the application:
//          Create the window, initialize data
//-----------------------------------------------------------------------------
function InitApp(Instance: THandle; nCmdShow: Longint): HResult;
var
    Wnd:    HWND;
    wc:      TWndClass;
    ddsd:    TDDSurfaceDesc2;
    ddscaps: TDDSCaps2;
    hRet:    HResult;
begin
    // Set up and register window class
    wc.style := CS_HREDRAW or CS_VREDRAW;
    wc.lpfnWndProc := @WindowProc;
    wc.cbClsExtra := 0;
    wc.cbWndExtra := 0;
    wc.hInstance := Instance;
    wc.hIcon := LoadIcon(Instance, MAKEINTRESOURCE(IDI_MAIN_ICON));
    wc.hCursor := LoadCursor(0, IDC_ARROW);
    wc.hbrBackground := GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName := NAME;
    wc.lpszClassName := NAME;
    RegisterClass(wc);

    // Create a window
    Wnd := CreateWindowEx(WS_EX_TOPMOST,
                          NAME,
                          TITLE,
                          WS_POPUP,
                          0,
                          0,
                          GetSystemMetrics(SM_CXSCREEN),
                          GetSystemMetrics(SM_CYSCREEN),
                          0,
                          0,
                          Instance,
                          nil);
    if Wnd = 0 then
    begin
        Result := LONGFALSE;
        exit;
    end;
    ShowWindow(Wnd, nCmdShow);
    UpdateWindow(Wnd);
    SetFocus(Wnd);

    ///////////////////////////////////////////////////////////////////////////
    // Create the main DirectDraw object
    ///////////////////////////////////////////////////////////////////////////
    hRet := DirectDrawCreateEx(nil, g_pDD, IID_IDirectDraw7, nil);
    if hRet <> DD_OK then
    begin
        Result := InitFail(Wnd, hRet, 'DirectDrawCreateEx FAILED');
        exit;
    end;

    // Get exclusive mode
    hRet := g_pDD^^.SetCooperativeLevel^(g_pDD, Wnd, DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN);
    if hRet <> DD_OK then
    begin
        Result := InitFail(Wnd, hRet, 'SetCooperativeLevel FAILED');
        exit;
    end;

    // Set the video mode to 640x480x8
    hRet := g_pDD^^.SetDisplayMode^(g_pDD, 640, 480, 8, 0, 0);
    if hRet <> DD_OK then
    begin
        Result := InitFail(Wnd, hRet, 'SetDisplayMode FAILED');
        exit;
    end;

    // Create the primary surface with 1 back buffer
    ZeroMemory(@ddsd, sizeof(ddsd));
    ddsd.dwSize := sizeof(ddsd);
    ddsd.dwFlags := DDSD_CAPS or DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or DDSCAPS_FLIP or
                           DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount := 1;
    hRet := g_pDD^^.CreateSurface^(g_pDD, ddsd, g_pDDSPrimary, nil);
    if hRet <> DD_OK then
    begin
        Result := InitFail(Wnd, hRet, 'CreateSurface FAILED');
        exit;
    end;

    // Get a pointer to the back buffer
    ZeroMemory(@ddscaps, sizeof(ddscaps));
    ddscaps.dwCaps := DDSCAPS_BACKBUFFER;
    hRet := g_pDDSPrimary^^.GetAttachedSurface^(g_pDDSPrimary, ddscaps, g_pDDSBack);
    if hRet <> DD_OK then
    begin
        Result := InitFail(Wnd, hRet, 'GetAttachedSurface FAILED');
        exit;
    end;

    // Create a offscreen bitmap.
    ddsd.dwFlags := DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
    ddsd.dwHeight := 480;
    ddsd.dwWidth := 640;
    hRet := g_pDD^^.CreateSurface^(g_pDD, ddsd, g_pDDSOne, nil);
    if (hRet <> DD_OK) then
    begin
        Result := InitFail(Wnd, hRet, 'CreateSurface FAILED');
        exit;
    end;

    // Create another offscreen bitmap.
    hRet := g_pDD^^.CreateSurface^(g_pDD, ddsd, g_pDDSTwo, nil);
    if (hRet <> DD_OK) then
    begin
        Result := InitFail(Wnd, hRet, 'CreateSurface FAILED');
        exit;
    end;

    // Create a Direct Draw Palette and associate it with the front buffer
    g_pDDPal := DDLoadPalette(g_pDD, szBitmap);
    if (g_pDDPal <> nil) then
        g_pDDSPrimary^^.SetPalette^(g_pDDSPrimary, g_pDDPal);
    if (not InitSurfaces) then
    begin
        Result := InitFail(Wnd, hRet, 'InitSurfaces FAILED');
        exit;
    end;

    // Create a timer to flip the pages
    if TIMER_ID <> SetTimer(Wnd, TIMER_ID, TIMER_RATE, nil) then
    begin
        Result := InitFail(Wnd, hRet, 'SetTimer FAILED');
        exit;
    end;

    Result := DD_OK;
end;




//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Initialization, message loop
//-----------------------------------------------------------------------------
var
   msg: TMsg;
begin
    if InitApp(hInstance, CmdShow) <> DD_OK then
        halt(0);

    while GetMessage(msg, 0, 0, 0) do
    begin
        TranslateMessage(msg);
        DispatchMessage(msg);
    end;
end.